@@ -4,7 +4,7 @@
  *	  WAL replay logic for inverted index.
  *
  *
- * Portions Copyright (c) 1996-2016, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2017, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -13,7 +13,9 @@
  */
 #include "postgres.h"
 
+#include "access/bufmask.h"
 #include "access/gin_private.h"
+#include "access/ginxlog.h"
 #include "access/xlogutils.h"
 #include "utils/memutils.h"
 
@@ -133,14 +135,6 @@ ginRedoInsertEntry(Buffer buffer, bool isLeaf, BlockNumber rightblkno, void *rda
 	}
 }
 
-/*
- * Redo recompression of posting list.  Doing all the changes in-place is not
- * always possible, because it might require more space than we've on the page.
- * Instead, once modification is required we copy unprocessed tail of the page
- * into separately allocated chunk of memory for further reading original
- * versions of segments.  Thanks to that we don't bother about moving page data
- * in-place.
- */
 static void
 ginRedoRecompress(Page page, ginxlogRecompressDataLeaf *data)
 {
@@ -150,9 +144,6 @@ ginRedoRecompress(Page page, ginxlogRecompressDataLeaf *data)
 	Pointer		segmentend;
 	char	   *walbuf;
 	int			totalsize;
-	Pointer		tailCopy = NULL;
-	Pointer		writePtr;
-	Pointer		segptr;
 
 	/*
 	 * If the page is in pre-9.4 format, convert to new format first.
@@ -192,7 +183,6 @@ ginRedoRecompress(Page page, ginxlogRecompressDataLeaf *data)
 	}
 
 	oldseg = GinDataLeafPageGetPostingList(page);
-	writePtr = (Pointer) oldseg;
 	segmentend = (Pointer) oldseg + GinDataLeafPageGetPostingListSize(page);
 	segno = 0;
 
@@ -210,6 +200,8 @@ ginRedoRecompress(Page page, ginxlogRecompressDataLeaf *data)
 		ItemPointerData *newitems;
 		int			nnewitems;
 		int			segsize;
+		Pointer		segptr;
+		int			szleft;
 
 		/* Extract all the information we need from the WAL record */
 		if (a_action == GIN_SEGMENT_INSERT ||
@@ -232,17 +224,6 @@ ginRedoRecompress(Page page, ginxlogRecompressDataLeaf *data)
 		Assert(segno <= a_segno);
 		while (segno < a_segno)
 		{
-			/*
-			 * Once modification is started and page tail is copied, we've
-			 * to copy unmodified segments.
-			 */
-			segsize = SizeOfGinPostingList(oldseg);
-			if (tailCopy)
-			{
-				Assert(writePtr + segsize < PageGetSpecialPointer(page));
-				memcpy(writePtr, (Pointer) oldseg, segsize);
-			}
-			writePtr += segsize;
 			oldseg = GinNextPostingListSegment(oldseg);
 			segno++;
 		}
@@ -283,42 +264,36 @@ ginRedoRecompress(Page page, ginxlogRecompressDataLeaf *data)
 			Assert(a_action == GIN_SEGMENT_INSERT);
 			segsize = 0;
 		}
-
-		/*
-		 * We're about to start modification of the page.  So, copy tail of the
-		 * page if it's not done already.
-		 */
-		if (!tailCopy && segptr != segmentend)
-		{
-			int tailSize = segmentend - segptr;
-
-			tailCopy = (Pointer) palloc(tailSize);
-			memcpy(tailCopy, segptr, tailSize);
-			segptr = tailCopy;
-			oldseg = (GinPostingList *) segptr;
-			segmentend = segptr + tailSize;
-		}
+		szleft = segmentend - segptr;
 
 		switch (a_action)
 		{
 			case GIN_SEGMENT_DELETE:
-				segptr += segsize;
+				memmove(segptr, segptr + segsize, szleft - segsize);
+				segmentend -= segsize;
+
 				segno++;
 				break;
 
 			case GIN_SEGMENT_INSERT:
+				/* make room for the new segment */
+				memmove(segptr + newsegsize, segptr, szleft);
 				/* copy the new segment in place */
-				Assert(writePtr + newsegsize <= PageGetSpecialPointer(page));
-				memcpy(writePtr, newseg, newsegsize);
-				writePtr += newsegsize;
+				memcpy(segptr, newseg, newsegsize);
+				segmentend += newsegsize;
+				segptr += newsegsize;
 				break;
 
 			case GIN_SEGMENT_REPLACE:
-				/* copy the new version of segment in place */
-				Assert(writePtr + newsegsize <= PageGetSpecialPointer(page));
-				memcpy(writePtr, newseg, newsegsize);
-				writePtr += newsegsize;
-				segptr += segsize;
+				/* shift the segments that follow */
+				memmove(segptr + newsegsize,
+						segptr + segsize,
+						szleft - segsize);
+				/* copy the replacement segment in place */
+				memcpy(segptr, newseg, newsegsize);
+				segmentend -= segsize;
+				segmentend += newsegsize;
+				segptr += newsegsize;
 				segno++;
 				break;
 
@@ -328,18 +303,7 @@ ginRedoRecompress(Page page, ginxlogRecompressDataLeaf *data)
 		oldseg = (GinPostingList *) segptr;
 	}
 
-	/* Copy the rest of unmodified segments if any. */
-	segptr = (Pointer) oldseg;
-	if (segptr != segmentend && tailCopy)
-	{
-		int restSize = segmentend - segptr;
-
-		Assert(writePtr + restSize <= PageGetSpecialPointer(page));
-		memcpy(writePtr, segptr, restSize);
-		writePtr += restSize;
-	}
-
-	totalsize = writePtr - (Pointer) GinDataLeafPageGetPostingList(page);
+	totalsize = segmentend - (Pointer) GinDataLeafPageGetPostingList(page);
 	GinDataPageSetDataSize(page, totalsize);
 }
 
@@ -511,42 +475,11 @@ ginRedoDeletePage(XLogReaderState *record)
 	Buffer		lbuffer;
 	Page		page;
 
-	/*
-	 * Lock left page first in order to prevent possible deadlock with
-	 * ginStepRight().
-	 */
-	if (XLogReadBufferForRedo(record, 2, &lbuffer) == BLK_NEEDS_REDO)
-	{
-		page = BufferGetPage(lbuffer);
-		Assert(GinPageIsData(page));
-		GinPageGetOpaque(page)->rightlink = data->rightLink;
-		PageSetLSN(page, lsn);
-		MarkBufferDirty(lbuffer);
-	}
-
 	if (XLogReadBufferForRedo(record, 0, &dbuffer) == BLK_NEEDS_REDO)
 	{
 		page = BufferGetPage(dbuffer);
 		Assert(GinPageIsData(page));
 		GinPageGetOpaque(page)->flags = GIN_DELETED;
-
-		/*
-		 * deleteXid field of ginxlogDeletePage was added during backpatching.
-		 * But, non-backpatched instances will continue generate WAL without
-		 * this field.  We should be able to correctly apply that.  We can
-		 * distinguish new WAL records by size their data, because
-		 * ginxlogDeletePage changes its size on both 32-bit and 64-bit
-		 * platforms.
-		 */
-		StaticAssertStmt(sizeof(ginxlogDeletePage) !=
-						 sizeof(ginxlogDeletePageOld),
-						 "ginxlogDeletePage size should be changed "
-						 "with addition of deleteXid field");
-		Assert(XLogRecGetDataLen(record) == sizeof(ginxlogDeletePage) ||
-			   XLogRecGetDataLen(record) == sizeof(ginxlogDeletePageOld));
-		if (XLogRecGetDataLen(record) == sizeof(ginxlogDeletePage))
-			GinPageSetDeleteXid(page, data->deleteXid);
-
 		PageSetLSN(page, lsn);
 		MarkBufferDirty(dbuffer);
 	}
@@ -561,6 +494,15 @@ ginRedoDeletePage(XLogReaderState *record)
 		MarkBufferDirty(pbuffer);
 	}
 
+	if (XLogReadBufferForRedo(record, 2, &lbuffer) == BLK_NEEDS_REDO)
+	{
+		page = BufferGetPage(lbuffer);
+		Assert(GinPageIsData(page));
+		GinPageGetOpaque(page)->rightlink = data->rightLink;
+		PageSetLSN(page, lsn);
+		MarkBufferDirty(lbuffer);
+	}
+
 	if (BufferIsValid(lbuffer))
 		UnlockReleaseBuffer(lbuffer);
 	if (BufferIsValid(pbuffer))
@@ -833,3 +775,34 @@ gin_xlog_cleanup(void)
 	MemoryContextDelete(opCtx);
 	opCtx = NULL;
 }
+
+/*
+ * Mask a GIN page before running consistency checks on it.
+ */
+void
+gin_mask(char *pagedata, BlockNumber blkno)
+{
+	Page		page = (Page) pagedata;
+	GinPageOpaque opaque;
+
+	mask_page_lsn_and_checksum(page);
+	opaque = GinPageGetOpaque(page);
+
+	mask_page_hint_bits(page);
+
+	/*
+	 * GIN metapage doesn't use pd_lower/pd_upper. Other page types do. Hence,
+	 * we need to apply masking for those pages.
+	 */
+	if (opaque->flags != GIN_META)
+	{
+		/*
+		 * For GIN_DELETED page, the page is initialized to empty. Hence, mask
+		 * the page content.
+		 */
+		if (opaque->flags & GIN_DELETED)
+			mask_page_content(page);
+		else
+			mask_unused_space(page);
+	}
+}
