@@ -9,7 +9,7 @@
  * exist, though, because mmap'd shmem provides no way to find out how
  * many processes are attached, which we need for interlocking purposes.
  *
- * Portions Copyright (c) 1996-2016, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2017, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  * IDENTIFICATION
@@ -38,6 +38,7 @@
 #include "storage/ipc.h"
 #include "storage/pg_shmem.h"
 #include "utils/guc.h"
+#include "utils/pidfile.h"
 
 
 /*
@@ -70,26 +71,6 @@
 typedef key_t IpcMemoryKey;		/* shared memory key passed to shmget(2) */
 typedef int IpcMemoryId;		/* shared memory ID returned by shmget(2) */
 
-/*
- * How does a given IpcMemoryId relate to this PostgreSQL process?
- *
- * One could recycle unattached segments of different data directories if we
- * distinguished that case from other SHMSTATE_FOREIGN cases.  Doing so would
- * cause us to visit less of the key space, making us less likely to detect a
- * SHMSTATE_ATTACHED key.  It would also complicate the concurrency analysis,
- * in that postmasters of different data directories could simultaneously
- * attempt to recycle a given key.  We'll waste keys longer in some cases, but
- * avoiding the problems of the alternative justifies that loss.
- */
-typedef enum
-{
-	SHMSTATE_ANALYSIS_FAILURE,	/* unexpected failure to analyze the ID */
-	SHMSTATE_ATTACHED,			/* pertinent to DataDir, has attached PIDs */
-	SHMSTATE_ENOENT,			/* no segment of that ID */
-	SHMSTATE_FOREIGN,			/* exists, but not pertinent to DataDir */
-	SHMSTATE_UNATTACHED			/* pertinent to DataDir, no attached PIDs */
-} IpcMemoryState;
-
 
 unsigned long UsedShmemSegID = 0;
 void	   *UsedShmemSegAddr = NULL;
@@ -102,9 +83,8 @@ static void *AnonymousShmem = NULL;
 static void *InternalIpcMemoryCreate(IpcMemoryKey memKey, Size size);
 static void IpcMemoryDetach(int status, Datum shmaddr);
 static void IpcMemoryDelete(int status, Datum shmId);
-static IpcMemoryState PGSharedMemoryAttach(IpcMemoryId shmId,
-					 void *attachAt,
-					 PGShmemHeader **addr);
+static PGShmemHeader *PGSharedMemoryAttach(IpcMemoryKey key,
+					 IpcMemoryId *shmid);
 
 
 /*
@@ -214,29 +194,29 @@ InternalIpcMemoryCreate(IpcMemoryKey memKey, Size size)
 		errno = shmget_errno;
 		ereport(FATAL,
 				(errmsg("could not create shared memory segment: %m"),
-		  errdetail("Failed system call was shmget(key=%lu, size=%zu, 0%o).",
-					(unsigned long) memKey, size,
-					IPC_CREAT | IPC_EXCL | IPCProtection),
+				 errdetail("Failed system call was shmget(key=%lu, size=%zu, 0%o).",
+						   (unsigned long) memKey, size,
+						   IPC_CREAT | IPC_EXCL | IPCProtection),
 				 (shmget_errno == EINVAL) ?
 				 errhint("This error usually means that PostgreSQL's request for a shared memory "
-		 "segment exceeded your kernel's SHMMAX parameter, or possibly that "
+						 "segment exceeded your kernel's SHMMAX parameter, or possibly that "
 						 "it is less than "
 						 "your kernel's SHMMIN parameter.\n"
-		"The PostgreSQL documentation contains more information about shared "
+						 "The PostgreSQL documentation contains more information about shared "
 						 "memory configuration.") : 0,
 				 (shmget_errno == ENOMEM) ?
 				 errhint("This error usually means that PostgreSQL's request for a shared "
 						 "memory segment exceeded your kernel's SHMALL parameter.  You might need "
 						 "to reconfigure the kernel with larger SHMALL.\n"
-		"The PostgreSQL documentation contains more information about shared "
+						 "The PostgreSQL documentation contains more information about shared "
 						 "memory configuration.") : 0,
 				 (shmget_errno == ENOSPC) ?
 				 errhint("This error does *not* mean that you have run out of disk space.  "
 						 "It occurs either if all available shared memory IDs have been taken, "
 						 "in which case you need to raise the SHMMNI parameter in your kernel, "
-		  "or because the system's overall limit for shared memory has been "
+						 "or because the system's overall limit for shared memory has been "
 						 "reached.\n"
-		"The PostgreSQL documentation contains more information about shared "
+						 "The PostgreSQL documentation contains more information about shared "
 						 "memory configuration.") : 0));
 	}
 
@@ -308,46 +288,14 @@ IpcMemoryDelete(int status, Datum shmId)
 bool
 PGSharedMemoryIsInUse(unsigned long id1, unsigned long id2)
 {
-	PGShmemHeader *memAddress;
-	IpcMemoryState state;
-
-	state = PGSharedMemoryAttach((IpcMemoryId) id2, NULL, &memAddress);
-	if (memAddress && shmdt(memAddress) < 0)
-		elog(LOG, "shmdt(%p) failed: %m", memAddress);
-	switch (state)
-	{
-		case SHMSTATE_ENOENT:
-		case SHMSTATE_FOREIGN:
-		case SHMSTATE_UNATTACHED:
-			return false;
-		case SHMSTATE_ANALYSIS_FAILURE:
-		case SHMSTATE_ATTACHED:
-			return true;
-	}
-	return true;
-}
-
-/*
- * Test for a segment with id shmId; see comment at IpcMemoryState.
- *
- * If the segment exists, we'll attempt to attach to it, using attachAt
- * if that's not NULL (but it's best to pass NULL if possible).
- *
- * *addr is set to the segment memory address if we attached to it, else NULL.
- */
-static IpcMemoryState
-PGSharedMemoryAttach(IpcMemoryId shmId,
-					 void *attachAt,
-					 PGShmemHeader **addr)
-{
+	IpcMemoryId shmId = (IpcMemoryId) id2;
 	struct shmid_ds shmStat;
 	struct stat statbuf;
 	PGShmemHeader *hdr;
 
-	*addr = NULL;
-
 	/*
-	 * First, try to stat the shm segment ID, to see if it exists at all.
+	 * We detect whether a shared memory segment is in use by seeing whether
+	 * it (a) exists and (b) has any processes attached to it.
 	 */
 	if (shmctl(shmId, IPC_STAT, &shmStat) < 0)
 	{
@@ -357,15 +305,15 @@ PGSharedMemoryAttach(IpcMemoryId shmId,
 		 * exists.
 		 */
 		if (errno == EINVAL)
-			return SHMSTATE_ENOENT;
+			return false;
 
 		/*
-		 * EACCES implies we have no read permission, which means it is not a
-		 * Postgres shmem segment (or at least, not one that is relevant to
-		 * our data directory).
+		 * EACCES implies that the segment belongs to some other userid, which
+		 * means it is not a Postgres shmem segment (or at least, not one that
+		 * is relevant to our data directory).
 		 */
 		if (errno == EACCES)
-			return SHMSTATE_FOREIGN;
+			return false;
 
 		/*
 		 * Some Linux kernel versions (in fact, all of them as of July 2007)
@@ -376,54 +324,33 @@ PGSharedMemoryAttach(IpcMemoryId shmId,
 		 */
 #ifdef HAVE_LINUX_EIDRM_BUG
 		if (errno == EIDRM)
-			return SHMSTATE_ENOENT;
+			return false;
 #endif
 
 		/*
-		 * Otherwise, we had better assume that the segment is in use.  The
-		 * only likely case is (non-Linux, assumed spec-compliant) EIDRM,
-		 * which implies that the segment has been IPC_RMID'd but there are
-		 * still processes attached to it.
+		 * Otherwise, we had better assume that the segment is in use. The
+		 * only likely case is EIDRM, which implies that the segment has been
+		 * IPC_RMID'd but there are still processes attached to it.
 		 */
-		return SHMSTATE_ANALYSIS_FAILURE;
+		return true;
 	}
 
+	/* If it has no attached processes, it's not in use */
+	if (shmStat.shm_nattch == 0)
+		return false;
+
 	/*
 	 * Try to attach to the segment and see if it matches our data directory.
 	 * This avoids shmid-conflict problems on machines that are running
-	 * several postmasters under the same userid and port number.  (That would
-	 * not ordinarily happen in production, but it can happen during parallel
-	 * testing.  Since our test setups don't open any TCP ports on Unix, such
-	 * cases don't conflict otherwise.)
+	 * several postmasters under the same userid.
 	 */
 	if (stat(DataDir, &statbuf) < 0)
-		return SHMSTATE_ANALYSIS_FAILURE;	/* can't stat; be conservative */
+		return true;			/* if can't stat, be conservative */
+
+	hdr = (PGShmemHeader *) shmat(shmId, NULL, PG_SHMAT_FLAGS);
 
-	hdr = (PGShmemHeader *) shmat(shmId, attachAt, PG_SHMAT_FLAGS);
 	if (hdr == (PGShmemHeader *) -1)
-	{
-		/*
-		 * Attachment failed.  The cases we're interested in are the same as
-		 * for the shmctl() call above.  In particular, note that the owning
-		 * postmaster could have terminated and removed the segment between
-		 * shmctl() and shmat().
-		 *
-		 * If attachAt isn't NULL, it's possible that EINVAL reflects a
-		 * problem with that address not a vanished segment, so it's best to
-		 * pass NULL when probing for conflicting segments.
-		 */
-		if (errno == EINVAL)
-			return SHMSTATE_ENOENT; /* segment disappeared */
-		if (errno == EACCES)
-			return SHMSTATE_FOREIGN;	/* must be non-Postgres */
-#ifdef HAVE_LINUX_EIDRM_BUG
-		if (errno == EIDRM)
-			return SHMSTATE_ENOENT; /* segment disappeared */
-#endif
-		/* Otherwise, be conservative. */
-		return SHMSTATE_ANALYSIS_FAILURE;
-	}
-	*addr = hdr;
+		return true;			/* if can't attach, be conservative */
 
 	if (hdr->magic != PGShmemMagic ||
 		hdr->device != statbuf.st_dev ||
@@ -431,17 +358,16 @@ PGSharedMemoryAttach(IpcMemoryId shmId,
 	{
 		/*
 		 * It's either not a Postgres segment, or not one for my data
-		 * directory.
+		 * directory.  In either case it poses no threat.
 		 */
-		return SHMSTATE_FOREIGN;
+		shmdt((void *) hdr);
+		return false;
 	}
 
-	/*
-	 * It does match our data directory, so now test whether any processes are
-	 * still attached to it.  (We are, now, but the shm_nattch result is from
-	 * before we attached to it.)
-	 */
-	return shmStat.shm_nattch == 0 ? SHMSTATE_UNATTACHED : SHMSTATE_ATTACHED;
+	/* Trouble --- looks a lot like there's still live backends */
+	shmdt((void *) hdr);
+
+	return true;
 }
 
 #ifdef USE_ANONYMOUS_SHMEM
@@ -515,10 +441,10 @@ GetHugePageSize(Size *hugepagesize, int *mmap_flags)
 			FreeFile(fp);
 		}
 	}
-#endif   /* __linux__ */
+#endif							/* __linux__ */
 }
 
-#endif   /* MAP_HUGETLB */
+#endif							/* MAP_HUGETLB */
 
 /*
  * Creates an anonymous mmap()ed shared memory segment.
@@ -579,10 +505,10 @@ CreateAnonymousSegment(Size *size)
 				(errmsg("could not map anonymous shared memory: %m"),
 				 (mmap_errno == ENOMEM) ?
 				 errhint("This error usually means that PostgreSQL's request "
-					"for a shared memory segment exceeded available memory, "
-					 "swap space, or huge pages. To reduce the request size "
+						 "for a shared memory segment exceeded available memory, "
+						 "swap space, or huge pages. To reduce the request size "
 						 "(currently %zu bytes), reduce PostgreSQL's shared "
-					   "memory usage, perhaps by reducing shared_buffers or "
+						 "memory usage, perhaps by reducing shared_buffers or "
 						 "max_connections.",
 						 *size) : 0));
 	}
@@ -608,7 +534,7 @@ AnonymousShmemDetach(int status, Datum arg)
 	}
 }
 
-#endif   /* USE_ANONYMOUS_SHMEM */
+#endif							/* USE_ANONYMOUS_SHMEM */
 
 /*
  * PGSharedMemoryCreate
@@ -617,21 +543,25 @@ AnonymousShmemDetach(int status, Datum arg)
  * standard header.  Also, register an on_shmem_exit callback to release
  * the storage.
  *
- * Dead Postgres segments pertinent to this DataDir are recycled if found, but
- * we do not fail upon collision with foreign shmem segments.  The idea here
- * is to detect and re-use keys that may have been assigned by a crashed
- * postmaster or backend.
+ * Dead Postgres segments are recycled if found, but we do not fail upon
+ * collision with non-Postgres shmem segments.  The idea here is to detect and
+ * re-use keys that may have been assigned by a crashed postmaster or backend.
+ *
+ * makePrivate means to always create a new segment, rather than attach to
+ * or recycle any existing segment.
  *
  * The port number is passed for possible use as a key (for SysV, we use
- * it to generate the starting shmem key).
+ * it to generate the starting shmem key).  In a standalone backend,
+ * zero will be passed.
  */
 PGShmemHeader *
-PGSharedMemoryCreate(Size size, int port,
+PGSharedMemoryCreate(Size size, bool makePrivate, int port,
 					 PGShmemHeader **shim)
 {
 	IpcMemoryKey NextShmemSegID;
 	void	   *memAddress;
 	PGShmemHeader *hdr;
+	IpcMemoryId shmid;
 	struct stat statbuf;
 	Size		sysvsize;
 
@@ -659,20 +589,14 @@ PGSharedMemoryCreate(Size size, int port,
 	sysvsize = size;
 #endif
 
-	/*
-	 * Loop till we find a free IPC key.  Trust CreateDataDirLockFile() to
-	 * ensure no more than one postmaster per data directory can enter this
-	 * loop simultaneously.  (CreateDataDirLockFile() does not ensure that,
-	 * but prefer fixing it over coping here.)
-	 */
-	NextShmemSegID = 1 + port * 1000;
+	/* Make sure PGSharedMemoryAttach doesn't fail without need */
+	UsedShmemSegAddr = NULL;
 
-	for (;;)
+	/* Loop till we find a free IPC key */
+	NextShmemSegID = port * 1000;
+
+	for (NextShmemSegID++;; NextShmemSegID++)
 	{
-		IpcMemoryId shmid;
-		PGShmemHeader *oldhdr;
-		IpcMemoryState state;
-
 		/* Try to create new segment */
 		memAddress = InternalIpcMemoryCreate(NextShmemSegID, sysvsize);
 		if (memAddress)
@@ -680,71 +604,58 @@ PGSharedMemoryCreate(Size size, int port,
 
 		/* Check shared memory and possibly remove and recreate */
 
+		if (makePrivate)		/* a standalone backend shouldn't do this */
+			continue;
+
+		if ((memAddress = PGSharedMemoryAttach(NextShmemSegID, &shmid)) == NULL)
+			continue;			/* can't attach, not one of mine */
+
 		/*
-		 * shmget() failure is typically EACCES, hence SHMSTATE_FOREIGN.
-		 * ENOENT, a narrow possibility, implies SHMSTATE_ENOENT, but one can
-		 * safely treat SHMSTATE_ENOENT like SHMSTATE_FOREIGN.
+		 * If I am not the creator and it belongs to an extant process,
+		 * continue.
 		 */
-		shmid = shmget(NextShmemSegID, sizeof(PGShmemHeader), 0);
-		if (shmid < 0)
+		hdr = (PGShmemHeader *) memAddress;
+		if (hdr->creatorPID != getpid())
 		{
-			oldhdr = NULL;
-			state = SHMSTATE_FOREIGN;
-		}
-		else
-			state = PGSharedMemoryAttach(shmid, NULL, &oldhdr);
-
-		switch (state)
-		{
-			case SHMSTATE_ANALYSIS_FAILURE:
-			case SHMSTATE_ATTACHED:
-				ereport(FATAL,
-						(errcode(ERRCODE_LOCK_FILE_EXISTS),
-						 errmsg("pre-existing shared memory block (key %lu, ID %lu) is still in use",
-								(unsigned long) NextShmemSegID,
-								(unsigned long) shmid),
-						 errhint("Terminate any old server processes associated with data directory \"%s\".",
-								 DataDir)));
-				break;
-			case SHMSTATE_ENOENT:
-
-				/*
-				 * To our surprise, some other process deleted since our last
-				 * InternalIpcMemoryCreate().  Moments earlier, we would have
-				 * seen SHMSTATE_FOREIGN.  Try that same ID again.
-				 */
-				elog(LOG,
-					 "shared memory block (key %lu, ID %lu) deleted during startup",
-					 (unsigned long) NextShmemSegID,
-					 (unsigned long) shmid);
-				break;
-			case SHMSTATE_FOREIGN:
-				NextShmemSegID++;
-				break;
-			case SHMSTATE_UNATTACHED:
-
-				/*
-				 * The segment pertains to DataDir, and every process that had
-				 * used it has died or detached.  Zap it, if possible, and any
-				 * associated dynamic shared memory segments, as well.  This
-				 * shouldn't fail, but if it does, assume the segment belongs
-				 * to someone else after all, and try the next candidate.
-				 * Otherwise, try again to create the segment.  That may fail
-				 * if some other process creates the same shmem key before we
-				 * do, in which case we'll try the next key.
-				 */
-				if (oldhdr->dsm_control != 0)
-					dsm_cleanup_using_control_segment(oldhdr->dsm_control);
-				if (shmctl(shmid, IPC_RMID, NULL) < 0)
-					NextShmemSegID++;
-				break;
+			if (kill(hdr->creatorPID, 0) == 0 || errno != ESRCH)
+			{
+				shmdt(memAddress);
+				continue;		/* segment belongs to a live process */
+			}
 		}
 
-		if (oldhdr && shmdt(oldhdr) < 0)
-			elog(LOG, "shmdt(%p) failed: %m", oldhdr);
+		/*
+		 * The segment appears to be from a dead Postgres process, or from a
+		 * previous cycle of life in this same process.  Zap it, if possible,
+		 * and any associated dynamic shared memory segments, as well. This
+		 * probably shouldn't fail, but if it does, assume the segment belongs
+		 * to someone else after all, and continue quietly.
+		 */
+		if (hdr->dsm_control != 0)
+			dsm_cleanup_using_control_segment(hdr->dsm_control);
+		shmdt(memAddress);
+		if (shmctl(shmid, IPC_RMID, NULL) < 0)
+			continue;
+
+		/*
+		 * Now try again to create the segment.
+		 */
+		memAddress = InternalIpcMemoryCreate(NextShmemSegID, sysvsize);
+		if (memAddress)
+			break;				/* successful create and attach */
+
+		/*
+		 * Can only get here if some other process managed to create the same
+		 * shmem key before we did.  Let him have that one, loop around to try
+		 * next key.
+		 */
 	}
 
-	/* Initialize new segment. */
+	/*
+	 * OK, we created a new segment.  Mark it as created by this process. The
+	 * order of assignments here is critical so that another Postgres process
+	 * can't see the header as valid but belonging to an invalid PID!
+	 */
 	hdr = (PGShmemHeader *) memAddress;
 	hdr->creatorPID = getpid();
 	hdr->magic = PGShmemMagic;
@@ -804,8 +715,7 @@ void
 PGSharedMemoryReAttach(void)
 {
 	IpcMemoryId shmid;
-	PGShmemHeader *hdr;
-	IpcMemoryState state;
+	void	   *hdr;
 	void	   *origUsedShmemSegAddr = UsedShmemSegAddr;
 
 	Assert(UsedShmemSegAddr != NULL);
@@ -818,18 +728,14 @@ PGSharedMemoryReAttach(void)
 #endif
 
 	elog(DEBUG3, "attaching to %p", UsedShmemSegAddr);
-	shmid = shmget(UsedShmemSegID, sizeof(PGShmemHeader), 0);
-	if (shmid < 0)
-		state = SHMSTATE_FOREIGN;
-	else
-		state = PGSharedMemoryAttach(shmid, UsedShmemSegAddr, &hdr);
-	if (state != SHMSTATE_ATTACHED)
+	hdr = (void *) PGSharedMemoryAttach((IpcMemoryKey) UsedShmemSegID, &shmid);
+	if (hdr == NULL)
 		elog(FATAL, "could not reattach to shared memory (key=%d, addr=%p): %m",
 			 (int) UsedShmemSegID, UsedShmemSegAddr);
 	if (hdr != origUsedShmemSegAddr)
 		elog(FATAL, "reattaching to shared memory returned unexpected address (got %p, expected %p)",
 			 hdr, origUsedShmemSegAddr);
-	dsm_set_control_handle(hdr->dsm_control);
+	dsm_set_control_handle(((PGShmemHeader *) hdr)->dsm_control);
 
 	UsedShmemSegAddr = hdr;		/* probably redundant */
 }
@@ -866,7 +772,7 @@ PGSharedMemoryNoReAttach(void)
 	UsedShmemSegID = 0;
 }
 
-#endif   /* EXEC_BACKEND */
+#endif							/* EXEC_BACKEND */
 
 /*
  * PGSharedMemoryDetach
@@ -905,3 +811,31 @@ PGSharedMemoryDetach(void)
 	}
 #endif
 }
+
+
+/*
+ * Attach to shared memory and make sure it has a Postgres header
+ *
+ * Returns attach address if OK, else NULL
+ */
+static PGShmemHeader *
+PGSharedMemoryAttach(IpcMemoryKey key, IpcMemoryId *shmid)
+{
+	PGShmemHeader *hdr;
+
+	if ((*shmid = shmget(key, sizeof(PGShmemHeader), 0)) < 0)
+		return NULL;
+
+	hdr = (PGShmemHeader *) shmat(*shmid, UsedShmemSegAddr, PG_SHMAT_FLAGS);
+
+	if (hdr == (PGShmemHeader *) -1)
+		return NULL;			/* failed: must be some other app's */
+
+	if (hdr->magic != PGShmemMagic)
+	{
+		shmdt((void *) hdr);
+		return NULL;			/* segment belongs to a non-Postgres app */
+	}
+
+	return hdr;
+}
